@@ -1,12 +1,11 @@
 #!/usr/bin/perl -w
 #
-# gnats.pl - Gnats www perl interface script
+# Gnatsweb - web front-end to gnats
 #
-# Copyright 1998 - Matt Gerassimoff
+# Copyright 1998-1999 - Matt Gerassimoff
 # and Ken Cox <kenstir@senteinc.com>
 #
-# $Revision: 1.24 $
-# $Header: /system/u/kenstir/cvsroot/gnatsweb/gnatsweb.pl,v 1.24 1999/04/02 05:05:07 kenstir Exp $
+# $Id: gnatsweb.pl,v 2.1 1999/04/11 17:11:24 kenstir Exp $
 #
 
 #-----------------------------------------------------------------------------
@@ -50,6 +49,8 @@ $site_mailer = '/usr/sbin/sendmail -oi -t';
 #     string    - string is used by gnatsweb according to $reason
 #
 # example:
+#     See gnatsweb-site-sente.pl for an extended example.
+#
 #     sub site_callback {
 #         my($reason, @args) = @_;
 #         if ($reason eq 'sendpr_description') {
@@ -68,11 +69,10 @@ use Socket;
 use IO::Handle;
 use POSIX;
 
-use Text::Wrap qw(wrap $columns $tabstop fill);
-$columns = 500;
-$tabstop = 8;
+# get RCS tag as just a number
+($REVISION  = '$Revision: 2.1 $ ') =~ s/.Revision: (.*) ../$1/;
 
-($REVISION  = '$Revision: 1.24 $ ') =~ s/\$\Revision: (.*) \$/$1/;
+# width of text fields
 $textwidth = 60;
 
 # where to get help -- a web site with translated info documentation
@@ -313,7 +313,7 @@ sub client_init
 
 # to debug:
 #     local($client_cmd_debug) = 1;
-#     client_cmd(args...);
+#     client_cmd(...);
 sub client_cmd
 {
   my($cmd) = @_;
@@ -324,8 +324,9 @@ sub client_cmd
 
 sub sendpr
 {
-  page_start_html('Create PR');
-  page_heading('Create Problem Report', 1);
+  my $page = 'Create PR';
+  page_start_html($page);
+  page_heading($page, 'Create Problem Report', 1);
 
   # remove "all" from arrays
   shift(@category);
@@ -384,13 +385,16 @@ sub sendpr
 	$q->submit('cmd', 'submit'),
 	" or ",
 	$q->reset(-name=>'reset'),
-	$q->end_form(),
-	$q->end_html();
+	$q->end_form();
+
+  page_footer($page);
+  page_end_html($page);
 }
 
 sub submitnewpr
 {
-  page_start_html('Create PR Results');
+  my $page = 'Create PR Results';
+  page_start_html($page);
 
   my(@values, $key);
   my(%fields);
@@ -406,7 +410,7 @@ sub submitnewpr
   }
   if($fields{'email'} eq "")
   {
-    page_heading('Error');
+    page_heading($page, 'Error');
     print "<h3>Your problem report has not been sent.</h3>\n";
     print "You must fill in the E-Mail Address field.\n";
     return;
@@ -415,7 +419,7 @@ sub submitnewpr
   $text = unparsepr("send", %fields);
   if(!open(MAIL, "|$site_mailer"))
   {
-    page_heading('Error');
+    page_heading($page, 'Error');
     print "<h3>Error invoking $site_mailer</h3>";
     return;
   }
@@ -436,33 +440,37 @@ EOT
   #print $q->pre(CGI::escapeHTML($text));
   if(!close(MAIL))
   {
-    page_heading('Error');
+    page_heading($page, 'Error');
     print "<h3>Bad pipe to $site_mailer</h3>";
     exit;
   }
-  page_heading('Problem Report Sent');
+  page_heading($page, 'Problem Report Sent');
   print "Thank you for your report.  It will take a short while for
 your report to be processed.  When it is, you will receive
 an automated message about it, containing the Problem Report
 number, and the developer who has been assigned to
 investigate the problem.";
+
+  page_footer($page);
+  page_end_html($page);
 }
 
 sub view
 {
-  page_start_html('View PR');
-
   my($viewaudit, $tmp) = @_;
 
+  my $page = 'View PR';
+  page_start_html($page);
+
   # $pr must be 'local' to be available to site callback
   local($pr) = $q->param('pr');
   if(!$pr)
   {
-    page_heading('Error');
+    page_heading($page, 'Error');
     print "<h3>You must specify a problem report number</h3>";
     return;
   }
-  page_heading("View Problem Report: $pr", 1);
+  page_heading($page, "View Problem Report: $pr", 1);
 
   # %fields must be 'local' to be available to site callback
   local(%fields) = readpr($pr);
@@ -470,14 +478,15 @@ sub view
   print $q->start_form(),
 	$q->hidden('pr', $pr),
         $q->submit('cmd', 'edit');
-  #if(!$viewaudit)
-  #{
-  #  print $q->submit('cmd', 'view audit');
-  #}
+  if(!$viewaudit)
+  {
+    print " or ", $q->submit('cmd', 'view audit-trail');
+  }
   print $q->hr(),
         "<table>";
   foreach (@fieldnames)
   {
+    next if $_ eq 'Audit-Trail';
     my $val = CGI::escapeHTML($fields{$_});
     my $valign = '';
     if ($fieldnames{$_} & $MULTILINE)
@@ -492,39 +501,39 @@ sub view
   print "</table>",
         $q->hr(),
         $q->submit('cmd', 'edit');
-  #if(!$viewaudit)
-  #{
-  #  print $q->submit('cmd', 'view audit');
-  #}
+  if(!$viewaudit)
+  {
+    print " or ", $q->submit('cmd', 'view audit-trail');
+  }
   print $q->end_form();
 
-  # Allow site callback to print a postlude.
-  my $html = cb('view_postlude');
-  print $html if $html;
+  # Footer comes before the audit-trail.
+  page_footer($page);
 
-  #if($viewaudit)
-  #{
-  #  print "<h3>Audit Trail:</h3><br>\n",
-  #        $q->pre(CGI::escapeHTML($fields{'Audit-Trail'}));
-  #}
+  if($viewaudit)
+  {
+    print "<h3>Audit Trail:</h3>\n",
+          $q->pre(CGI::escapeHTML($fields{'Audit-Trail'}));
+  }
 
-  print $q->end_html();
+  page_end_html($page);
 }
 
 sub edit
 {
-  page_start_html('Edit PR');
+  my $page = 'Edit PR';
+  page_start_html($page);
 
   my $debug = 0;
 
   my($pr) = $q->param('pr');
   if(!$pr)
   {
-    page_heading('Error');
+    page_heading($page, 'Error');
     print "<h3>You must specify a problem report number</h3>";
     return;
   }
-  page_heading("Edit Problem Report: $pr", 1);
+  page_heading($page, "Edit Problem Report: $pr", 1);
 
   my(%fields) = readpr($pr);
 #  if ($debug)
@@ -599,22 +608,27 @@ sub edit
 	" or ",
 	$q->reset(-name=>'reset'),
 	$q->end_form(),
-	$q->hr(),
-	"<h3>Audit-Trail:</h3>\n",
-	$q->pre(CGI::escapeHTML($fields{'Audit-Trail'})),
-	$q->end_html();
+	$q->hr();
+
+  # Footer comes before the audit-trail.
+  page_footer($page);
+
+  print "<h3>Audit-Trail:</h3>\n",
+        $q->pre(CGI::escapeHTML($fields{'Audit-Trail'}));
+  page_end_html($page);
 }
 
 sub submitedit
 {
-  page_start_html('Edit PR Results');
+  my $page = 'Edit PR Results';
+  page_start_html($page);
 
   my $debug = 0;
 
   my($pr) = $q->param('pr');
   if(!$pr)
   {
-    page_heading('Error');
+    page_heading($page, 'Error');
     print "<h3>ERROR: You must specify a problem report number</h3>";
     return;
   }
@@ -660,7 +674,7 @@ sub submitedit
 
     if($fields{'Last-Modified'} ne $oldfields{'Last-Modified'})
     {
-      page_heading('Error');
+      page_heading($page, 'Error');
       print "<h3>Sorry, PR $pr has been modified since you started editing it.</h3>\n",
             "Please return to the edit form, press the Reload button, ",
             "then make your edits again.\n";
@@ -671,7 +685,7 @@ sub submitedit
 
     if($user eq "" || $fields{'Responsible'} eq "")
     {
-      page_heading('Error');
+      page_heading($page, 'Error');
       print "<h3>Responsible party is '$fields{'Responsible'}', user is '$user'!</h3>\n";
       last LOCKED;
     }
@@ -702,8 +716,8 @@ sub submitedit
 	  {
 	    if($fields{"$_-Why"} =~ /^\s*$/)
 	    {
-              page_heading('Error') if $ok;
-	      print "<h3>Field $_ must have a reason for change</h3>";
+              page_heading($page, 'Error') if $ok;
+	      print "<h3>Field '$_' must have a reason for change</h3>";
 	      $ok = 0;
 	    }
             else
@@ -731,7 +745,7 @@ sub submitedit
       {
 	if(!defined($adr = praddr($fields{'Responsible'})))
 	{
-          page_heading('Error');
+          page_heading($page, 'Error');
 	  print "<h3>Cannot find address for $fields{'Responsible'}.</h3>";
 	  last LOCKED;
 	}
@@ -739,7 +753,7 @@ sub submitedit
       }
       if(!defined($adr = praddr($oldfields{'Responsible'})))
       {
-        page_heading('Error');
+        page_heading($page, 'Error');
 	print "<h3>Cannot find address for $oldfields{'Responsible'}.</h3>";
 	last LOCKED;
       }
@@ -782,7 +796,7 @@ sub submitedit
           print MAILER $q->url() . "?cmd=view&pr=$fields{'Number'}\n";
           if(!close(MAILER))
 	  {
-            page_heading('Error');
+            page_heading($page, 'Error');
 	    print "<h3>Edit successful, but email notification failed</h3>",
                   "Bad pipe to $site_mailer";
 	    last LOCKED;
@@ -791,13 +805,13 @@ sub submitedit
 	}
 	else
 	{
-          page_heading('Error');
+          page_heading($page, 'Error');
           print "<h3>Edit successful, but email notification failed</h3>",
                 "Can't run $site_mailer";
           last LOCKED;
 	}
       }
-      page_heading('Edit Successful');
+      page_heading($page, 'Edit Successful');
       print "<h3>Your changes to PR $fields{'Number'} were filed to the database.</h3>\n";
       print "The parties concerned were notified via e-mail:<br>",
             "<pre>$mailto</pre>"
@@ -805,12 +819,16 @@ sub submitedit
     }
   }
   unlockpr($fields{'Number'});
+
+  page_footer($page);
+  page_end_html($page);
 }
 
 sub query_page
 {
-  page_start_html('Query PR');
-  page_heading('Query Problem Reports', 1);
+  my $page = 'Query PR';
+  page_start_html($page);
+  page_heading($page, 'Query Problem Reports', 1);
   print_stored_queries();
   print $q->start_form(),
 	"<table>",
@@ -860,14 +878,17 @@ sub query_page
 	"</table>",
         "<br><br>",
 	$q->submit('cmd', 'submit query'),
-        $q->end_form(),
-        $q->end_html();
+        $q->end_form();
+
+  page_footer($page);
+  page_end_html($page);
 }
 
 sub advanced_query_page
 {
-  page_start_html('Advanced Query');
-  page_heading('Query Problem Reports', 1);
+  my $page = 'Advanced Query';
+  page_start_html($page);
+  page_heading($page, 'Query Problem Reports', 1);
   print_stored_queries();
   print $q->start_form();
 
@@ -983,14 +1004,16 @@ sub advanced_query_page
   print "</center>\n";
   print "<hr>",
 	$q->submit('cmd', 'submit query'),
-	$q->end_form(),
-	$q->end_html();
+	$q->end_form();
+  page_footer($page);
+  page_end_html($page);
 }
 
 sub submitquery
 {
-  page_start_html('Query Results');
-  page_heading('Query Results', 1);
+  my $page = 'Query Results';
+  page_start_html($page);
+  page_heading($page, 'Query Results', 1);
   my $debug = 0;
 
   my $originatedbyme = $q->param('originatedbyme');
@@ -1006,11 +1029,15 @@ sub submitquery
   foreach $param ($q->param())
   {
     next unless $gnatsd_query{$param};
-    # Turn multiple values into regular expression.
+
+    # Turn multiple param values into regular expression.
     @val = $q->param($param);
     $regexp = join('|', @val);
-    # Discard trailing '|all'.
+
+    # Discard trailing '|all', or leading '|'.
     $regexp =~ s/\|all$//;
+    $regexp =~ s/^\|//;
+
     # If there's still a query here, make it.
     client_cmd("$gnatsd_query{$param} $regexp")
           if($regexp && $regexp ne 'all');
@@ -1019,6 +1046,8 @@ sub submitquery
   my(@query_results) = client_cmd("sql2");
 
   display_query_results(@query_results);
+  page_footer($page);
+  page_end_html($page);
 }
 
 # by_field -
@@ -1088,9 +1117,11 @@ sub display_query_results
   my(@fields) = $q->param('columns');
   my($field, %fields);
 
-  #print $q->h1("Query Results:", scalar(@query_results), "matches");
   my $num_matches = scalar(@query_results);
-  print $q->h2($num_matches ? $num_matches : "No", "matches found");
+  my $heading = sprintf("%s %s found",
+                        $num_matches ? $num_matches : "No",
+                        ($num_matches == 1) ? "match" : "matches");
+  print $q->h2($heading);
 
   # Sort @query_results according to the rules in by_field().
   # Using the "map, sort" idiom allows us to perform the expensive
@@ -1189,7 +1220,6 @@ sub display_query_results
         $q->submit('cmd', 'store query'),
         "</table>",
         $q->end_form();
-  print $q->end_html();
 }
 
 # store_query -
@@ -1207,20 +1237,25 @@ sub store_query
   $q->delete('cmd');
   my $query_string = $q->query_string();
 
+  # Have to generate the cookie before printing the header.
   my $new_cookie = $q->cookie(-name => "gnatsweb-query-$queryname",
                               -value => $query_string,
                               -expires => '+10y');
   print $q->header(-cookie => $new_cookie);
-  page_start_html('Query Saved');
-  page_heading('Query Saved');
+
+  # Now print the page.
+  my $page = 'Query Saved';
+  page_start_html($page);
+  page_heading($page, 'Query Saved');
   print "<h2>debugging</h2><pre>",
         "query_string: $query_string",
         "cookie: $new_cookie\n",
         "</pre><hr>\n"
         if $debug;
   print "<p>Your query has been saved.  It will be available ",
-        "the next time you reload the Query page.",
-        $q->end_html();
+        "the next time you reload the Query page.";
+  page_footer($page);
+  page_end_html($page);
 }
 
 # print_stored_queries -
@@ -1281,7 +1316,8 @@ sub submit_stored_query
   {
     print $q->header(),
           $q->start_html('Error'),
-          $q->h3($err);
+          $q->h3($err),
+          $q->end_html();
   }
   else
   {
@@ -1302,8 +1338,9 @@ sub submit_stored_query
 
 sub help_page
 {
-  page_start_html('Help');
-  page_heading('Help', 1);
+  my $page = 'Help';
+  page_start_html($page);
+  page_heading($page, 'Help', 1);
   print p('Welcome to our problem report database. ',
           'You\'ll notice that here we call them "problem reports" ',
           'or "PR\'s", not "bugs".');
@@ -1313,7 +1350,8 @@ sub help_page
           a({-href=>"$gnats_info_top?(gnats)"}, 'the gnats manual'),
           'to read about bug lifecycles and the like, ',
           'but then again, you may not.');
-  print $q->end_html();
+  page_footer($page);
+  page_end_html($page);
 }
 
 sub one_line_form
@@ -1335,8 +1373,9 @@ sub can_edit
 
 sub main_page
 {
-  page_start_html('Main');
-  page_heading('Main Page', 1);
+  my $page = 'Main';
+  page_start_html($page);
+  page_heading($page, 'Main Page', 1);
   print '<p><table>';
 
   one_line_form('Create Problem Report:',
@@ -1365,12 +1404,14 @@ sub main_page
   one_line_form('Get Help:',
                 submit('cmd', 'help'));
   print '</table>';
+  page_footer($page);
   print '<hr><small>',
-        $q->escapeHTML('Gnatsweb 2.0 brought to you by ' .
+        "Gnatsweb $REVISION, brought to you by<br>",
+        $q->escapeHTML(
                        'Matt Gerassimoff <mg@digalogsys.com> ' .
                        'and Kenneth H. Cox <kenstir@senteinc.com>.'),
         '</small>';
-  print $q->end_html();
+  page_end_html($page);
 }
 
 # cb -
@@ -1495,10 +1536,10 @@ END_OF_STYLE
 #
 sub page_heading
 {
-  my($heading, $display_user_info) = @_;
+  my($title, $heading, $display_user_info) = @_;
 
   # Allow site callback to override html.
-  my $html = cb('page_heading', $heading);
+  my $html = cb('page_heading', $title, $heading);
   if ($html)
   {
     print $html;
@@ -1522,6 +1563,39 @@ sub page_heading
                  td({-nowrap=>1}, $rightcol)));
 }
 
+# page_footer -
+#
+#     Allow the site_callback to take control before the end of the
+#     page.
+#
+sub page_footer
+{
+  my $title = shift;
+
+  my $html = cb('page_footer', $title);
+  print $html if ($html);
+}
+
+# page_end_html -
+#
+#     Print the HTML which ends a page.  Allow the site_callback to
+#     take control here too.
+#
+sub page_end_html
+{
+  my $title = shift;
+
+  # Allow site callback to override html.
+  my $html = cb('page_end_html', $title);
+  if ($html)
+  {
+    print $html;
+    return;
+  }
+
+  print $q->end_html();
+}
+
 # fix_multiline_val -
 #     Modify text of multitext field so that it contains \n separators
 #     (not \r\n or \n as some platforms use), and so that it has a \n
@@ -1807,8 +1881,9 @@ sub praddr
 
 sub login
 {
-  page_start_html('Login');
-  page_heading('Login');
+  my $page = 'Login';
+  page_start_html($page);
+  page_heading($page, 'Login');
 
   client_init();
   my(@dbs) = client_cmd("dbla");
@@ -1829,8 +1904,9 @@ sub login
                        -default=>$database),
         "</table>",
         $q->submit('cmd','login'),
-        $q->end_form(),
-	$q->end_html();
+        $q->end_form();
+  page_footer($page);
+  page_end_html($page);
 }
 
 #
@@ -1930,6 +2006,10 @@ sub main
     submitnewpr();
   }
   elsif($cmd eq 'view')
+  {
+    view(0);
+  }
+  elsif($cmd eq 'view audit-trail')
   {
     view(1);
   }
